;; Generated package description from shelisp.el  -*- no-byte-compile: t -*-
(define-package "shelisp" "0.9.1.0.20210926.72746" "execute elisp in shell" 'nil :url "https://elpa.gnu.org/packages/shelisp.html" :authors '(("Michael R. Mauger" . "michael@mauger.com")) :maintainer '("Michael R. Mauger" . "michael@mauger.com") :keywords '("terminals" "lisp" "processes"))
